"""Tests

Can run all tests in all files by running this from root of TermHub:
    python -m unittest discover
"""
import os
import sys
import unittest
from pathlib import Path
from typing import List, Set, Tuple

import pandas as pd

TEST_DIR = Path(os.path.abspath(os.path.dirname(__file__)))
TEST_INPUT_DIR = TEST_DIR / 'input'
TEST_OUTPUT_DIR = TEST_DIR / 'output'
# TODO: Delete temp vars when not needed
TEMP_CONCEPT_CSV = '/Users/joeflack4/projects/TermHub/termhub-csets/datasets/prepped_files/concept.csv'
TEMP_CONCEPT_REL_CSV = '/Users/joeflack4/projects/TermHub/termhub-csets/datasets/prepped_files/concept_relationship.csv'
PROJECT_ROOT = TEST_DIR.parent
sys.path.insert(0, str(PROJECT_ROOT))
from omop2owl_vocab import CONCEPT_DTYPES, CONCEPT_RELATIONSHIP_DTYPES
from omop2fhir_vocab.omop2fhir_vocab import omop2fhir


# TODO: _create_test_files() and _prep_combine_test_subsets() duplicated from omop2owl. how to DRY?
def _create_test_files(
    concept_csv_path: str = TEMP_CONCEPT_CSV, concept_relationship_csv_path: str = TEMP_CONCEPT_REL_CSV,
):
    """Create test files"""
    # Read inputs
    # - concept table
    concept_df = pd.read_csv(concept_csv_path, dtype=CONCEPT_DTYPES).fillna('')
    # todo: del index_col line
    # concept_df = pd.read_csv(concept_csv_path, index_col='concept_id', dtype=CONCEPT_DTYPES).fillna('')

    # - concept_relationship table
    concept_rel_df = pd.read_csv(concept_relationship_csv_path, dtype=CONCEPT_RELATIONSHIP_DTYPES).fillna('')
    concept_rel_df = concept_rel_df[concept_rel_df.invalid_reason == '']

    vocabs = ['ICD10CM', 'SNOMED', 'RxNorm', 'NDC', 'CPT4']  # arbitrary
    # vocabs = [x for x in list(concept_df['vocabulary_id'].unique()) if x]  # all
    for voc in vocabs:
        concept_df_i = concept_df[concept_df.vocabulary_id == voc]
        concept_df_i2 = concept_df_i.head(20)  # 20 = arbitrary
        concept_ids: Set[str] = set(concept_df_i2.concept_id)
        concept_rel_df_i = concept_rel_df[
            (concept_rel_df.concept_id_1.isin(concept_ids)) |
            (concept_rel_df.concept_id_2.isin(concept_ids))]
        this_test_dir = TEST_INPUT_DIR / voc
        os.makedirs(this_test_dir, exist_ok=True)
        concept_df_i2.to_csv(str(this_test_dir / 'concept.csv'), index=False)
        concept_rel_df_i.to_csv(str(this_test_dir / 'concept_relationship.csv'), index=False)

# TODO: add back unittest superclass
# class TestOmop2Fhir(unittest.TestCase):
class TestOmop2Fhir():
    """Tests"""

    @staticmethod
    def _prep_combine_test_subsets(use_cache=True, create_fresh_test_files=False) -> Tuple[Path, Path]:
        """The tests are set up to be able to run snippets of vocabs, so as a pre-step, joins them."""
        # Vars
        outdir = TEST_OUTPUT_DIR / 'combined_inputs'
        concept_outpath = outdir / 'concept.csv'
        concept_rel_outpath = outdir / 'concept_relationship.csv'
        test_vocs: List[str] = os.listdir(TEST_INPUT_DIR)

        # Create test files
        if create_fresh_test_files:
            _create_test_files()

        # Use cache
        if use_cache and os.path.exists(concept_outpath) and os.path.exists(concept_rel_outpath):
            return concept_outpath, concept_rel_outpath

        # Combine
        os.makedirs(outdir, exist_ok=True)
        concept_dfs: List[pd.DataFrame] = []
        concept_rel_dfs: List[pd.DataFrame] = []
        for test_voc in test_vocs:
            concept_dfs.append(pd.read_csv(TEST_INPUT_DIR / test_voc / 'concept.csv'))
            concept_rel_dfs.append(pd.read_csv(TEST_INPUT_DIR / test_voc / 'concept_relationship.csv'))
        concept_df = pd.concat(concept_dfs)
        concept_rel_df = pd.concat(concept_rel_dfs)

        # Save & return
        concept_df.to_csv(concept_outpath, index=False)
        concept_rel_df.to_csv(concept_rel_outpath, index=False)
        return concept_outpath, concept_rel_outpath

    def test_defaults_except_all_rels(self):
        """Test default settings, except for including all relationships"""
        # Vars
        concept_outpath, concept_rel_outpath = self._prep_combine_test_subsets()
        out_dir = TEST_OUTPUT_DIR / 'test_defaults_except_all_rels'
        settings = {
            'concept_csv_path': str(concept_outpath),
            'concept_relationship_csv_path': str(concept_rel_outpath),
            'out_dir': str(out_dir),
            'relationships': ['ALL'],  # default
            # 'relationships': ['Is a'],
            # vocabs: List[str] = [],
            # exclude_vocab_prefix: False,
            # TODO: temporary until fixed: https://github.com/mapping-commons/sssom-py/issues/456
            'include_conceptmap': False,
        }
        # Run program
        for file in os.listdir(out_dir):
            path = out_dir / file
            if not os.path.isdir(path):
                os.remove(path)
        omop2fhir(**settings)

        # Tests: CodeSystem
        # TODO
        # todo: read JSON and check
        # ids = []
        # rels = []
        # rel_set = set([x[1] for x in rels])
        # self.assertGreater(len(ids), 100)
        # self.assertGreater(len(rels), 50)
        # self.assertIn('rdfs:subClassOf', rel_set)

        # Tests: ConceptMap
        # TODO
        print()


# todo: print: for some reason stuff is not printing. When I run tests in another project, they do print. Supposedly
#  it is not uncommon (default?) for things not to print though. id like to fix if i can, so I can see progress
# TODO: add back unittest superclass
# Special debugging: To debug in PyCharm and have it stop at point of error, change TestOmop2Owl(unittest.TestCase)
#  to TestOmop2Fhir, and uncomment below.
if __name__ == '__main__':
    tester = TestOmop2Fhir()
    tester.test_defaults_except_all_rels()



"""omop2fhir-vocab: Convert OMOP vocab to FHIR."""
from datetime import datetime

from omop2fhir_vocab.omop2fhir_vocab import cli


if __name__ == '__main__':
    t1 = datetime.now()
    cli()
    t2 = datetime.now()
    print(f'Finished in {(t2 - t1).seconds} seconds')



"""Setup for packaging"""
import io
import os
import sys
from setuptools import find_packages, setup, Command
from shutil import rmtree


PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = 'omop2fhir-vocab'
DESCRIPTION = 'Convert OMOP vocab to FHIR.'
URL = 'https://github.com/HOT-Ecosystem/omop2fhir-vocab'
EMAIL = 'jflack@jhu.edu'
AUTHOR = 'Joe Flack'
REQUIRES_PYTHON = '>=3.9.0'
VERSION = '1.1.2'

# Requirements
REQUIRED = [
    'omop2owl-vocab>=1.3.4',
    'owl-on-fhir>=1.1.1',
]

# Description
with io.open(os.path.join(PROJECT_ROOT, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()


# As of 3.5 I believe, this doesn't work. Use `make pypi` and `make pypi-test`, which use twine.
class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        """Init options"""
        pass

    def finalize_options(self):
        """Finalize options"""
        pass

    def run(self):
        """Run upload"""
        self.status('Removing previous builds…')
        dist_dir = os.path.join(PROJECT_ROOT, 'dist')
        if os.path.exists(dist_dir):
            rmtree(dist_dir)
        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))
        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')
        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')
        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('test',)),
    package_data={},
    install_requires=REQUIRED,
    # extras_require=EXTRAS,
    include_package_data=True,
    # license='MIT',  # todo: add LICENSE.md from GitHub and add license
    classifiers=[
        # todo: does this really qualify?
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
    entry_points={
        'console_scripts': [
            'omop2fhir-vocab = omop2fhir_vocab.__main__:cli'
        ]
    },
)



"""omop2fhir-vocab: Convert OMOP vocab to FHIR."""
import os
from argparse import ArgumentParser
from datetime import datetime
from typing import Dict, List

# noinspection PyProtectedMember
from omop2owl_vocab.omop2owl_vocab import omop2owl
from owl_on_fhir.__main__ import owl_to_fhir

PROG = 'omop2fhir-vocab'
DESC = 'Convert OMOP vocab to FHIR.'
CACHE_OPTIONS = ['all', 'omop2fhir', 'omop2owl', 'owl2fhir']


def omop2fhir(
    concept_csv_path: str, concept_relationship_csv_path: str, vocabs: List[str] = [],
    relationships: List[str] = ['Is a'], out_dir: str = None,
    vocab_outputs=True, combined_outputs=True, exclude_vocab_prefix=False,
    exclude_codesystem=False, exclude_conceptmap=False, caching: List[str] = [],
    # Non CLI flags:
    skip_semsql=True
):
    """Run program"""
    # OMOP2OWL
    print('--- Serializaing to intermediate format: OWL ---')
    omop2owl_cache = True if 'all' in caching or 'omop2owl' in caching else False
    d: Dict = omop2owl(
        concept_csv_path, concept_relationship_csv_path, vocabs=vocabs, relationships=relationships, outdir=out_dir,
        skip_semsql=skip_semsql, retain_general_cache=omop2owl_cache, use_cache=omop2owl_cache)

    # Manipulations
    d2 = {}
    if not exclude_vocab_prefix:
        for mode, mode_d in d.items():
            d2[mode] = {}
            for vocab_id, path in mode_d.items():
                d2[mode][f'OMOP-{vocab_id}'] = path
    d = d2

    # OWL2FHIR
    print('--- Converting to FHIR ---')
    # todo: code_system_url? (if not created automatically)
    owl2fhir_cache = True if 'all' in caching or 'owl2fhir' in caching else False
    try:
        if vocab_outputs:
            for vocab_id, path in d['vocab_outputs'].items():
                print(f'- converting vocab: "{vocab_id}"')
                owl_to_fhir(
                    path, out_dir, code_system_id=vocab_id, include_codesystem=not exclude_codesystem,
                    include_conceptmap=not exclude_conceptmap,
                    retain_intermediaries=owl2fhir_cache, use_cached_intermediaries=owl2fhir_cache)
        if combined_outputs:
            print('- creating combined output')
            _id, path = [x for x in d['combined_output'].items()][0]
            owl_to_fhir(
                path, out_dir, code_system_id=_id, include_codesystem=not exclude_codesystem,
                include_conceptmap=not exclude_conceptmap,
                retain_intermediaries=owl2fhir_cache, use_cached_intermediaries=owl2fhir_cache)
    # - teardown
    finally:
        if not 'all' in caching and not 'omop2fhir' in caching:
            for mode in d.values():
                for path in mode.values():
                    os.remove(path)


def cli_parser(title: str = PROG, description: str = DESC) -> ArgumentParser:
    """Get CLI parser"""
    parser = ArgumentParser(prog=title, description=description)

    # omop2owl-vocab only args -----------------------------------------------------------------------------------------
    # Required
    parser.add_argument(
        '-c', '--concept-csv-path', required=True, help='Path to CSV of OMOP concept table.')
    parser.add_argument(
        '-r', '--concept-relationship-csv-path', required=True,
        help='Path to CSV of OMOP concept_relationship table.')
    # Optional
    parser.add_argument(
        '-v', '--vocabs', required=False, nargs='+',
        help='Which vocabularies to include in the output?  Usage: --vocabs "Procedure Type" "Device Type"')
    parser.add_argument(
        '-R', '--relationships', required=False, nargs='+', default=['Is a'],
        help='Which relationship types from the concept_relationship table\'s relationship_id field to include? '
             'Default is "Is a" only. Passing "ALL" includes everything. Usage: --realationships "Is a" "Maps to"')

    # Common arguments -------------------------------------------------------------------------------------------------
    # todo: mark anywhere where verbiage differs
    #  - Ideally, while doing so, fix it so that owl2fhir and omop2owl have same verbiage
    parser.add_argument(
        '-o', '--out-dir', required=False, default=os.getcwd(),
        help='Output directory. Defaults to current working directory.')
    # parser.add_argument(  # omop2owl version
    #     '-O', '--outdir', required=False, default=os.getcwd(), help='Output directory.')
    # - not implemented
    #   - caching: it is parameterized differently in eaach package
    # todo: consider adding: codesys-id/ontology-id: very similar but may be worth 2 params
    # parser.add_argument(
    #     '-s', '--code-system-id', required=True, default=False,
    #     help="The code system ID to use for identification on the server uploaded to. "
    #          "See: https://hl7.org/fhir/resource-definitions.html#Resource.id")
    # # parser.add_argument(  # omop2owl version. should this be exposed separately?
    #     '-I', '--ontology-id', required=False, default='OMOP',  # add str(randint(100000, 999999))?
    #     help='Identifier for ontology. Used to generate a pURL and file name.')
    # - todo: expose any of these?
    # parser.add_argument(
    #     '-C', '--use-cache', required=False, action='store_true',
    #     help='Of outputs or intermediates already exist, use them.')
    # parser.add_argument(  # I think this is useful for omop2owl because of semsql, but we don't use that output here
    #     '-M', '--memory', required=False, default=100, help='The amount of Java memory (GB) to allocate.')
    # parser.add_argument(
    #     '-o', '--output-type', required=False, default='merged-post-split',
    #     choices=['merged', 'split', 'merged-post-split', 'rxnorm'],
    #     help='What output to generate? If "merged" will create an ONTOLOGY_ID.db file with all concepts of all vocabs '
    #          'merged into one. If "split" will create an ONTOLOGY_ID-*.db file for each vocab. "merged-post-split" '
    #          'output will be as if running both "split" and  "merged", but the merging implementation is different. '
    #          'Use this option if running out of memory. If using "rxnorm", will create a specifically customized '
    #          'ONTOLOGY_ID-RxNorm.db.')
    # - not useful
    # parser.add_argument(
    #     '-S', '--skip-semsql', required=False, action='store_true',
    #     help='In addition to .owl, also convert to a SemanticSQL .db? This is always True except when --output-type is '
    #          'all-merged-post-split and it is creating initial .owl files to be merged.')
    # parser.add_argument(
    #     '-e', '--exclude-singletons', required=False, action='store_true',
    #     help='Exclude terms that do not have any relationships. This only applies to --method robot.')
    # parser.add_argument(
    #     '-s', '--semsql-only', required=False, action='store_true',
    #     help='Use this if the .owl already exists and you just want to create a SemanticSQL .db.')

    # New args ---------------------------------------------------------------------------------------------------------
    parser.add_argument(
        '-V', '--vocab-outputs', action='store_true', default=True, required=False,
        help='Create set of artefacts (CodeSystem, ConceptMap) for each vocabulary?')
    parser.add_argument(
        '-C', '--combined-outputs', action='store_true', default=True, required=False,
        help='Create set of artefacts (CodeSystem, ConceptMap) for all of OMOP vocabulary combined?')
    parser.add_argument(
        '-e', '--exclude-vocab-prefix', action='store_true', default=False, required=False,
        help='With this flag absent, the identifiers and file names for each vocabulary will be prefixed "OMOP".')
    parser.add_argument(
        '-S', '--exclude-code-system', action='store_true', default=False, required=False,
        help='Exclude CodeSystem outputs.')
    parser.add_argument(
        '-M', '--exclude-concept-map', action='store_true', default=False, required=False,
        help='Exclude ConceptMap outputs.')
    parser.add_argument(
        '-H', '--caching', nargs='+', required=False, choices=CACHE_OPTIONS,
        # TODO: Needs improvement! Careful when using. Read warning below.
        help='Caching options. Warning: Developer option. Use at your own risk. Currently this option is in development'
             ' and it will only work if you don\'t change your options after running it for the first time. If you use '
             'this and then want to change your options, particularly --vocabs and --relationships, you will have to '
             '(a) not use caching for subsequent runs, or (b) uninstall and reinstall the program. '
             '`all`: Turns on all cache options. `omop2fhir`: Saves any intermediate files at the top '
             'level of this pipeline, which basically entails caching of output from omop2owl which is used by owl2fhir'
             '. `omop2owl`: Intermediates used by that package, such as robot templates. `owl2fhir`: Intermediates used'
             ' by that package, such as Obographs JSON.')

    # owl2fhir only args -----------------------------------------------------------------------------------------------
    # - todo: expose any of these?
    # parser.add_argument(
    #     '-S', '--code-system-url', required=True, default=False,
    #     help="Canonical URL for the code system. "
    #          "See: https://hl7.org/fhir/codesystem-definitions.html#CodeSystem.url")
    # parser.add_argument(
    #     '-t', '--intermediary-type', choices=INTERMEDIARY_TYPES, default='obographs', required=False,
    #     help='Which type of intermediary to use? First, we convert OWL to that intermediary format, and then we '
    #          'convert that to FHIR.')
    # parser.add_argument(
    #     '-c', '--use-cached-intermediaries', action='store_true', required=False, default=False,
    #     help='Use cached intermediaries if they exist? Also will save intermediaries to owl-on-fhir\'s cache/ dir.')
    # parser.add_argument(
    #     '-r', '--retain-intermediaries', action='store_true', default=False, required=False,
    #     help='Retain intermediary files created during conversion process (e.g. Obograph JSON)?')
    # parser.add_argument(
    #     '-I', '--convert-intermediaries-only', action='store_true', default=False, required=False,
    #     help='Convert intermediaries only?')
    # - not useful
    # parser.add_argument('-i', '--input-path-or-url', required=True, help='URL or path to OWL file to convert.')
    # parser.add_argument(  # covered by omop2owl --relationships
    #     '-p', '--include-only-critical-predicates', action='store_true', required=False, default=False,
    #     help='If present, includes only critical predicates (is_a/parent) rather than all predicates in '
    #          'CodeSystem.property and CodeSystem.concept.property.')
    # parser.add_argument(
    #     '-d', '--dev-oak-path', default=False, required=False,
    #     help='If you want to use a local development version of OAK, specify the path to the OAK directory here. '
    #          'Must be used with --dev-oak-interpreter-path.')
    # parser.add_argument(
    #     '-D', '--dev-oak-interpreter-path', default=False, required=False,
    #     help='If you want to use a local development version of OAK, specify the path to the Python interpreter where '
    #          'its dependencies are installed (i.e. its virtual environment). Must be used with --dev-oak-path.')
    # parser.add_argument(
    #     '-u', '--native-uri-stems', required=True, nargs='+',
    #     help='A comma-separated list of URI stems that will be used to determine whether a concept is native to '
    #          'the CodeSystem. For example, for OMIM, the following URI stems are native: '
    #          'https://omim.org/entry/,https://omim.org/phenotypicSeries/PS". '
    #          'As of 2023-01-15, there is still a bug in the Obographs spec and/or `robot` where certain nodes are not'
    #          ' being converted. This converter adds back the nodes, but to know which ones belong to the CodeSystem '
    #          'itself and are not foreign concepts, this parameter is necessary. OAK also makes use of this parameter. '
    #          'See also: https://github.com/geneontology/obographs/issues/90')
    return parser


def cli(title: str = PROG, description: str = DESC):
    """Command line interface.

    This package is mostly a wrapper around other packages. As such, arguments will be passed down."""
    parser: ArgumentParser = cli_parser(title, description)
    d: Dict = vars(parser.parse_args())
    omop2fhir(**d)


if __name__ == '__main__':
    t1 = datetime.now()
    cli()
    t2 = datetime.now()
    print(f'Finished in {(t2 - t1).seconds} seconds')
